#ifndef _PANEL_H
extern "C" {
#include <panel.h>
}
#define _PANEL_H
#endif

#include "AUI.h"

static void
putCommand (WINDOW *win, int row, int col, shInterface::Command cmd)
{
    wattrset (win, ColorMap[kBrightMagenta]);
    mvwaddstr (win, row, col, I->getKeyForCommand (cmd));
    wattrset (win, A_NORMAL);
}

static void
putCommandL (WINDOW *win, int row, shInterface::Command cmd)
{
    putCommand (win, row, 1, cmd);
}

static void
putCommandR (WINDOW *win, int row, shInterface::Command cmd)
{
    putCommand (win, row, 41, cmd);
}

void
shASCIIInterface::showHelp ()
{
    WINDOW *win;
    PANEL *panel;
    int row;
    int has_vi_keys = mKey2Cmd['l'] == kMoveE;
    int has_vi_fire_keys = mKey2Cmd[0x1f & 'l'] == kFireE;

    int height, width;
    getmaxyx(stdscr, height, width);

    win = newwin (height, 80, 0, 0);
    if (!win) {
        I->p ("Couldn't create help window.");
        return;
    }
    panel = new_panel (win);

    wattrset (win, ColorMap[kBrightMagenta]);
    mvwaddstr (win, 0, 2, "7 8 9");
    mvwaddstr (win, 2, 2, "4   6");
    mvwaddstr (win, 4, 2, "1 2 3");

    wattrset (win, A_NORMAL);
    mvwaddstr (win, 1, 2, " \\|/");
    mvwaddstr (win, 2, 3,  "-@-");
    mvwaddstr (win, 3, 2, " /|\\");

    mvwaddstr (win, 0, 9, "Use the numeric keypad or arrow keys for movement.");
    mvwaddstr (win, 1, 9, "Move into a monster to fight it in melee.");
    mvwaddstr (win, 2, 9, "Use ");
    wattrset (win, ColorMap[kBrightMagenta]);
    waddstr (win, getKeyForCommand (kGlide));
    wattrset (win, A_NORMAL);
    waddstr (win, " key followed by a direction key to move in that direction");
    mvwaddstr (win, 3, 9, "until something interesting is found.  Rest with ");
    wattrset (win, ColorMap[kBrightMagenta]);
    waddstr (win, "5");
    wattrset (win, A_NORMAL);
    waddstr (win, " key.");
    mvwaddstr (win, 4, 9, "Use ");
    wattrset (win, ColorMap[kBrightMagenta]);
    waddstr (win, getKeyForCommand (kMoveUp));
    wattrset (win, A_NORMAL);
    waddstr (win, " and ");
    wattrset (win, ColorMap[kBrightMagenta]);
    waddstr (win, getKeyForCommand (kMoveDown));
    wattrset (win, A_NORMAL);
    waddstr (win, " to select respectively up and down directions.");

    if (has_vi_keys) {
        wattrset (win, ColorMap[kBrightMagenta]);
        mvwaddstr (win, 6, 2,  "y k u  Y K U");
        mvwaddstr (win, 8, 2, "h   l  H   L");
        mvwaddstr (win, 10, 2, "b j n  B J N");
        wattrset (win, A_NORMAL);
        mvwaddstr (win, 7, 2,  " \\|/    \\|/");
        mvwaddstr (win, 8, 3,  "-@-");
        mvwaddstr (win, 8, 10, "-@-");
        mvwaddstr (win, 9, 2, " /|\\    /|\\");

        mvwaddstr (win, 6, 16, "Nethack/vi movement keys are also supported");
        mvwaddstr (win, 7, 16, "by selected keymap. ");
        wattrset (win, ColorMap[kBrightMagenta]);
        waddstr (win, "YUHJKLBN");
        wattrset (win, A_NORMAL);
        waddstr (win, " will move in");
        mvwaddstr (win, 8, 16, "the direction until interesting is found.");
        if (has_vi_fire_keys) {
            wattrset (win, ColorMap[kBrightMagenta]);
            mvwaddstr (win, 9, 16, "Control");
            wattrset (win, A_NORMAL);
            waddstr (win, " + ");
            wattrset (win, ColorMap[kBrightMagenta]);
            waddstr (win, "YUHJKLBN");
            wattrset (win, A_NORMAL);
            waddstr (win, " shoots in the direction.");
        }
        mvwaddstr (win, 10, 16, "Additionally ");
        wattrset (win, ColorMap[kBrightMagenta]);
        waddstr (win, ".");
        wattrset (win, A_NORMAL);
        waddstr (win, " key will also rest.");
    }

    row = 6 + 6 * has_vi_keys;
    putCommandL (win, row, kAdjust);
    mvwaddstr (win, row++, 8, "Adjust shortcut letter");
    putCommandL (win, row, kExamine);
    mvwaddstr (win, row++, 8, "Examine object closely");
    putCommandL (win, row, kUse);
    mvwaddstr (win, row++, 8, "Apply or activate (use)");
    putCommandL (win, row, kQuaff);
    mvwaddstr (win, row++, 8, "Drink object contents (quaff)");
    putCommandL (win, row, kExecute);
    mvwaddstr (win, row++, 8, "Execute (run object's program)");
    putCommandL (win, row, kZapRayGun);
    mvwaddstr (win, row++, 8, "Zap a ray gun");
    putCommandL (win, row, kName);
    mvwaddstr (win, row++, 8, "Name an object or its class");
    putCommandL (win, row, kWield);
    mvwaddstr (win, row++, 8, "Hold something in hands (wield)");
    putCommandL (win, row, kWear);
    mvwaddstr (win, row++, 8, "Wear armor, install implants");
    putCommandL (win, row, kTakeOff);
    mvwaddstr (win, row++, 8, "Take off armor, remove implants");
    putCommandL (win, row, kThrow);
    mvwaddstr (win, row++, 8, "Throw an object");
    putCommandL (win, row, kFireWeapon);
    mvwaddstr (win, row++, 8, "Fire wielded weapon");
    putCommandL (win, row, kKick);
    mvwaddstr (win, row++, 8, "Kick a door, monster or object");

    row = 6 + 6 * has_vi_keys;
    putCommandR (win, row, kListInventory);
    mvwaddstr (win, row++, 48, "Browse your inventory");
    putCommandR (win, row, kPickup);
    mvwaddstr (win, row++, 48, "Pick up items from the floor");
    putCommandR (win, row, kPickUpAll);
    mvwaddstr (win, row++, 48, "Pick up whole stack of items");
    putCommandR (win, row, kDrop);
    mvwaddstr (win, row++, 48, "Drop an item");
    putCommandR (win, row, kDropMany);
    mvwaddstr (win, row++, 48, "Drop several items");
    putCommandR (win, row, kPay);
    mvwaddstr (win, row++, 48, "Pay for an item or service");
    putCommandR (win, row, kMutantPower);
    mvwaddstr (win, row++, 48, "Use a mutant power");
    putCommandR (win, row, kHistory);
    mvwaddstr (win, row++, 48, "Show console message history");
    putCommandR (win, row, kLook);
    mvwaddstr (win, row++, 48, "Look at a feature or monster");
    putCommandR (win, row, kOpen);
    mvwaddstr (win, row++, 48, "Open a door, operate droids");
    putCommandR (win, row, kClose);
    mvwaddstr (win, row++, 48, "Close a door");
    putCommandR (win, row, kDiscoveries);
    mvwaddstr (win, row++, 48, "Show recognized item types");
    putCommandR (win, row, kEditSkills);
    mvwaddstr (win, row++, 48, "Open skill screen");

#if 0
        if (i < 0x1f) {     /* ctrl */
            mvwaddch (win, row, 3, '^' | ColorMap[kBrightMagenta]);
            key = i | 0x40;
        } else if (0x80 == (i & 0x80)) {    /* alt */
            mvwaddch (win, row, 2, 'M' | ColorMap[kBrightMagenta]);
            mvwaddch (win, row, 3, '-' | ColorMap[kBrightMagenta]);
            key = i & 0x70;
        }
#endif
    update_panels ();
    doupdate();
    getChar ();
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
    drawScreen ();
}
