#ifndef OBJECTTYPE_H
#define OBJECTTYPE_H

//#include <stdlib.h>

#include "Attack.h"
#include "ObjectIlk.h"

/* Various types of objects in the game.
   ObjectType is the most general classification of an object,
   ObjectIlk is a more specific subclassification.
*/

extern shGlyph ObjectGlyphs[];
extern int ColorMap[22];

enum shThingSize
{                 /* creature size */
    kFine,        /* 6 in or less */
    kDiminutive,
    kTiny,
    kSmall,
    kMedium,
    kLarge,
    kHuge,
    kGigantic,
    kColossal
};

enum shMaterialType  //borrowed from Nethack
{
    kLiquid,
    kWax,
    kVegetable,
    kFleshy,
    kPaper,
    kCloth,
    kPlastic,
    kLeather,
    kWood,
    kBone,
    kIron,
    kBrass,
    kTin,
    kBronze,
    kLead,
    kSteel,
    kAluminum,
    kTitanium,
    kCarbonFiber,
    kPlasteel,
    kAdamantium,
    kEndurium,
    kDepletedUranium,
    kCopper,
    kSilver,
    kElectrum,
    kGold,
    kPlatinum,
    kMithril,
    kSilicon,
    kGlass,
    kCrystal,
    kGem,
    kMineral,
    kRock
};


enum shIntrinsics {
    kTelepathy =           0x1,
    kMotionDetection =     0x2,
    kTremorsense =         0x4,
    kScent =               0x8,
    kXRayVision =          0x10,
    kReflection =          0x20,
    kAutoSearching =       0x40,
    kTranslation =         0x80,  /* babel fish */
    kHealthMonitoring =    0x100, /* low HP warning */
    kRadiationProcessing = 0x200,
    kSenseLife =           0x400, /* Xel'Naga racial ability */
    kAutoRegeneration =    0x800,
    kLightSource =         0x1000,
    kPerilSensing =        0x2000,
    kBlind =               0x4000,
    kShielded =            0x8000, /* protected by a shield generator */
    kBugSensing =          0x10000, /* software engineer */
    kFlying =              0x20000,
    kLucky =               0x40000,
    kNarcolepsy =          0x80000,
    kBrainShielded =       0x100000,
    kScary =               0x200000,
    kAcidBlood =           0x400000, /* splash damage on death */
    kMultiplier =          0x800000,
    kAirSupply =           0x1000000, /* has air supply */
    kBreathless =          0x2000000, /* doesn't need to breathe */
    kCanSwim =             0x4000000,
    kNightVision =         0x8000000,
    kCamouflaged =         0x10000000, /* Invisible next to obstacles. */
    kInvisible =           0x20000000, /* Covered by cloaking field. */
    kEMFieldVision =       0x40000000  /* Like night vision but for aliens. */
};


enum shObjectIlkFlags {
    kMergeable = 0x1,
    kChargeable = 0x2,
    kIdentified = 0x4,
    kEnhanceable = 0x8,
    kBugProof = 0x10,
    kUsuallyBuggy = 0x20,
    kUnique = 0x40,  /* Use the instead of a/an indefinite article. */
    kIndestructible = 0x80,
    kRadioactive = 0x100, /* Always emits radiation. */
    kPowered = 0x200, /* The weight of worn powered armor doesn't encumber. */
    kSelectiveFire = 0x400, /* Gun has burst and single shot modes. */
    kSealed = 0x800, /* Protects against vacuum. */
    kPreidentified = 0x1004  /* Do not show in discoveries table. */
};
/* NOTE: If you want to add something that emits radiation only
   under certain conditions do it through shObject::isRadioactive
   and shObject::isRadioactiveKnown instead.  The flag is for
   objects that irradiate surroundings no matter what. */


struct shObjectDescData {
    char mDesc[40];
    shColor mColor;
};

extern shObjectDescData CanisterData[30];
extern shObjectDescData ImplantData[30];
extern shObjectDescData FloppyData[30];
extern shObjectDescData RayGunData[30];
extern shObjectDescData JumpsuitData[30];
extern shObjectDescData BeltData[5];
extern shObjectDescData KeycardData[5];
extern shObjectDescData GizmoData[5];
extern shObjectDescData GrenadeData[30];
extern shObjectDescData SGrenadeData[30];

void randomizeIlkNames ();
void randomizeServiceNames ();

enum shExecResult
{
    kNormal = 0x0,
    kNoExpiry = 0x1,
    kNoEat = 0x2,
    kDestruct = 0x4, /* destruct */
    kFakeExpiry = 0x8, /* always pretend license ended and then destruct */
    kGoneAlready = 0x10, /* assume the pointer is invalid */
    kDestroyComputer = 0x20 /* delete computer and floppy disk */
};

/* following return ms elapsed */
typedef int shToolFunc (shObject *tool);
typedef int shCanisterFunc (shObject *canister);
typedef int shCanisterUseFunc (shObject *canister);
typedef shExecResult shFloppyDiskFunc (shObject *computer, shObject *disk);

struct shObjectIlk
{
    /* base properties */
    shObjId mId;
    shObjectType mType;
    shObjId mParent; // parent ilk, useful for isA ()
                     // e.g. bullet is parent of silver bullet

    const char *mName;       /* precise name, e.g. "canister of restoration */
    const char *mAppearance; /* ignorant description, e.g. "green canister" */
    const char *mVagueName;  /* blind name, e.g. "canister" */
    const char *mUserName;   /* name given by user, e.g.: "tastes like crap" */
    shGlyph mGlyph;

    shMaterialType mMaterial;

    int mCost;         /* base cost in shop */

    int mFlags;
    int mCarriedIntrinsics; /* conferred to the carrier */
    int mWornIntrinsics;    /* conferred only when worn */
    int mWieldedIntrinsics; /* conferred only when wielded */
    int mActiveIntrinsics;  /* conferred only when active */

    int mProbability; /* relative probability when randomly generating;
                         the constant ABSTRACT indicates an abstract ilk */

    int mWeight;       // in grams
    shThingSize mSize; // physical size / unwieldiness.
    int mHardness;     // flat incoming damage reduction
    int mHP;           // base HP
    int mEnergyUse;    /* for items that continuously use energy (e.g. geiger
                          counters, flashlights), ticks per cell consumed
                          for items that use energy in bursts (e.g. weapons),
                          cells consumed per use */

    /* weapon properties */
    int mRange;   // in ft, 0 indicates melee weapon
    int mThreatRange;
    int mCriticalMultiplier;
    int mToHitBonus;
    shObjId mAmmoType;
    int mAmmoBurst; /* no of rounds of ammo /energy cells consumed per burst */
    shSkillCode mMeleeSkill;
    shSkillCode mGunSkill;
    shAttack *mMeleeAttack;   /* bump with this wielded */
    shAttack *mMissileAttack; /* throw it */
    shAttack *mGunAttack;     /* fire it */

    /* ray gun properties */
    shAttack *mZapAttack;     /* zap it */

    /* implant properties */
    enum Site {
        kFrontalLobe,
        kParietalLobe,
        kOccipitalLobe,
        kTemporalLobe,
        kCerebellum,
        kLeftEar,
        kRightEyeball,
        kNeck,

        kMaxSite,
        kAnyEar,
        kAnyEye,
        kAnyBrain,
        kNoSite
    };
    Site mSite;
    shSkillCode mBoostedSkill;

    /* armor properties */
    char mArmorBonus;
    unsigned char mMaxEnhancement;
    signed char mResistances[kMaxEnergyType]; /* resistances granted by this armor */

    int mEquipSpeed;  //ms to don or doff armor
    unsigned char mSpeedPenalty; //in % - 100% penalty == movement takes twice as long
    int mToHitModifier;
    int mDamageModifier;
    int mSpeedBoost;

    /* any worn or wielded */
    int mPsiModifier;

    /* tool properties */
    int mMaxCharges;
    shToolFunc *mUseFunc;

    /* canister properties */
    shCanisterFunc *mQuaffFunc;

    /* floppy disk properties */
    shFloppyDiskFunc *mRunFunc;

    /* association with monster type */
    //shMonsterIlk *mMonsterIlk;

    /* leftover stuff */
    int isA (shObjId ilk);
    char *getRayGunColor (); /* kludgey, not reentrant*/
    shSkillCode getSkillForAttack (shAttack *attack);
    shObjectIlk
    (
        shObjId id,
        shObjectType type,
        shObjId parent,
        const char *name, const char *appearance, const char *vagueName,
        shColor color, // used to compose default glyph
        shMaterialType material,
        int cost, int flags,
        int carriedIntrinsics, int wornIntrinsics,
        int wieldedIntrinsics, int activeIntrinsics,
        int prob,
        int weight, shThingSize size,
        int hardness, int hp,
        int energyUse,
        int toHitBonus,
        shObjId ammoType,
        int ammoBurst,
        shSkillCode mskill, shSkillCode gskill,
        shAttack *meleeAttack,
        shAttack *missileAttack,
        shAttack *gunAttack,
        shAttack *zapAttack,
        Site site,
        shSkillCode boostedSkill,
        char armorBonus,
        unsigned char maxEnhancement,
        int equipSpeed,
        unsigned char speedPenalty,
        int toHitModifier,
        int damageModifier,
        int speedBoost,
        int psiModifier,
        shEnergyType energy,
        int resist,
        int maxCharges,
        shToolFunc *useFunc,
        shCanisterFunc *quaffFunc,
        shFloppyDiskFunc* runFunc
    )
    {
        mId = id;
        mParent = parent;
        mType = type;
        mFlags = flags;
        mGlyph.mSym = ObjectGlyphs[type].mSym;
        mGlyph.mColor = color;
        mMaterial = material;
        mCost = cost; mProbability = prob;
        mCarriedIntrinsics = carriedIntrinsics; mWornIntrinsics = wornIntrinsics;
        mWieldedIntrinsics = wieldedIntrinsics; mActiveIntrinsics = activeIntrinsics;
        mName = name; mAppearance = appearance; mVagueName = vagueName;
        mUserName = NULL;
        mWeight = weight; mSize = size;
        mHardness = hardness; mHP = hp;
        mEnergyUse = energyUse;
        mToHitBonus = toHitBonus;
        mAmmoType = ammoType;
        mAmmoBurst = ammoBurst;
        mMeleeSkill = mskill;
        mGunSkill = gskill;
        mMeleeAttack = meleeAttack;
        mMissileAttack = missileAttack;
        mGunAttack = gunAttack;
        mZapAttack = zapAttack;
        if (mGunAttack) {
            mRange = mGunAttack->mRange;
        } else if (mMissileAttack) {
            mRange = mMissileAttack->mRange;
        } else {
            mRange = 0; /* Melee only */
        }
        mSite = site;
        mBoostedSkill = boostedSkill;
        mArmorBonus = armorBonus;
        mMaxEnhancement = maxEnhancement;
        mEquipSpeed = equipSpeed;
        mSpeedPenalty = speedPenalty;
        mToHitModifier = toHitModifier; mDamageModifier = damageModifier;
        mSpeedBoost = speedBoost;
        mPsiModifier = psiModifier;
        for (int i = 0; i < kMaxEnergyType; ++i) {
            mResistances[i] = 0;
        }
        mResistances[energy] = resist;
        mMaxCharges = maxCharges;
        mUseFunc = useFunc; mQuaffFunc = quaffFunc; mRunFunc = runFunc;
    }
};

const char *describeImplantSite (shObjectIlk::Site site);

int makeRepair (shObject *tool);

#define NOINTRINSICS 0, 0, 0, 0
#define NOUSAGES NULL, NULL, NULL
#define NOT_A_WEAPON \
    0, kObjNothing, 0, kNoSkillCode, kNoSkillCode, NULL, NULL, NULL, NULL
#define EXPLOSIVE(ATTACK) \
    0, kObjNothing, 0, kNoSkillCode, kNoSkillCode, NULL, ATTACK, NULL, NULL
#define NO_AMMO \
    kObjNothing, 0
#define RAY_GUN(ATTACK) \
    0, kObjNothing, 0, kNoSkillCode, kNoSkillCode, NULL, NULL, NULL, ATTACK
#define NOT_AN_IMPLANT \
    shObjectIlk::kNoSite, kNoSkillCode
#define NOT_AN_ARMOR \
    0, 0, 0, 0, 0, 0, 0, 0, kNoEnergy, 0
#define NOT_AN_ARMOR_PSI(PSI) \
    0, 0, 0, 0, 0, 0, 0, PSI, kNoEnergy, 0

#define NEW_FLOPPY_DISK(enum, NAME, LABEL, COLOR, FUNC, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kFloppyDisk, kObjGenericFloppyDisk, NAME, LABEL, "floppy disk",\
    COLOR, kPlastic, COST, kMergeable, NOINTRINSICS, RARITY, 18, kFine,\
    3, 1, 0, NOT_A_WEAPON, NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NULL, NULL, FUNC\
)

#define NEW_CANISTER(enum, NAME, LABEL, COLOR, FLAGS, QUAFFFUNC, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kCanister, kObjGenericCanister, NAME, LABEL, "canister",\
    COLOR, kGlass, COST, FLAGS | kMergeable, NOINTRINSICS, RARITY, 400, kTiny,\
    1, 1, 0, NOT_A_WEAPON, NOT_AN_IMPLANT, NOT_AN_ARMOR, 0,\
    useCanister, QUAFFFUNC, NULL\
)

#define NEW_EXPLOSIVE_CANISTER(enum, NAME, LABEL, COLOR, FLAGS, ATTACK,\
    QUAFFFUNC, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kCanister, kObjGenericCanister, NAME, LABEL, "canister",\
    COLOR, kGlass, COST, FLAGS | kMergeable, NOINTRINSICS, RARITY, 400, kTiny,\
    1, 1, 0, EXPLOSIVE(ATTACK), NOT_AN_IMPLANT, NOT_AN_ARMOR, 0,\
    useCanister, QUAFFFUNC, NULL\
)

#define NEW_RAY_GUN(enum, NAME, LABEL, COLOR, FLAGS, ATTACK, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kRayGun, kObjGenericRayGun, NAME, LABEL, "ray gun", COLOR, kSteel,\
    COST, FLAGS | kChargeable, NOINTRINSICS, RARITY, 500, kMedium,\
    10, 3, 0, RAY_GUN(ATTACK), NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_IMPLANT(enum, NAME1, NAME2, NAME3, COLOR, PARENT, FLAGS, SITE,\
    INTRINSICS, SKILL, PSIMOD, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kImplant, PARENT, NAME1, NAME3, NAME2, COLOR, kFleshy, COST,\
    FLAGS, 0, INTRINSICS, 0, 0, RARITY, 20, kDiminutive, 5, 1, 0, NOT_A_WEAPON,\
    SITE, SKILL, NOT_AN_ARMOR_PSI(PSIMOD), 0, NOUSAGES\
)

#define NEW_ARMOR(enum, NAME1, NAME2, NAME3, COLOR, PARENT, MATERIAL,\
    FLAGS, WEIGHT, SIZE, HARDNESS, HP, PSIMOD, ARMOR, MAXENHANCE, RESISTTYPE,\
    RESISTVALUE, EQUIPTIME, MOVEPENALTY, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kArmor, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST, FLAGS,\
    NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0, NOT_A_WEAPON,\
    NOT_AN_IMPLANT, ARMOR, MAXENHANCE, EQUIPTIME, MOVEPENALTY, 0, 0, 0,\
    PSIMOD, RESISTTYPE, RESISTVALUE, 0, NOUSAGES\
)

#define NEW_TOOL(enum, NAME1, NAME2, NAME3, COLOR, PARENT, COST, MATERIAL,\
    FLAGS, WEIGHT, SIZE, HARDNESS, HP, ENERGYUSE, MAXCHARGES, FUNC, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kTool, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST, FLAGS,\
    NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, ENERGYUSE,\
    NOT_A_WEAPON, NOT_AN_IMPLANT, NOT_AN_ARMOR, MAXCHARGES, FUNC, NULL, NULL\
)

#define NEW_WEAPON_TOOL(enum, NAME1, NAME2, NAME3, COLOR, PARENT, COST,\
    MATERIAL, FLAGS, WEIGHT, SIZE, HARDNESS, HP, MELEE, ENERGYUSE, MAXCHARGES,\
    FUNC, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kTool, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST, FLAGS,\
    NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, ENERGYUSE, +0,\
    kObjNothing, 0, kMeleeWeapon, kNoSkillCode, MELEE, NULL, NULL, NULL,\
    NOT_AN_IMPLANT, NOT_AN_ARMOR, MAXCHARGES, FUNC, NULL, NULL\
)

#define NEW_GENERIC_WEAPON(enum, NAME1, NAME2, NAME3, COLOR, MSKILL, GSKILL,\
    PARENT, MATERIAL, FLAGS, WEIGHT, SIZE, HARDNESS, HP, TOHIT, MELEE, MISSILE,\
    GUN, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kWeapon, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kEnhanceable, NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0,\
    TOHIT, kObjNothing, 0, MSKILL, GSKILL, MELEE, MISSILE, GUN, NULL,\
    NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_MELEE_WEAPON(enum, NAME1, NAME2, NAME3, COLOR, MSKILL, PARENT,\
    MATERIAL, FLAGS, WEIGHT, SIZE, HARDNESS, HP, TOHIT, MELEE, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kWeapon, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kEnhanceable, NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0,\
    TOHIT, kObjNothing, 0, MSKILL, kNoSkillCode, MELEE, NULL, NULL, NULL,\
    NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_POWERED_MELEE_WEAPON(enum, NAME1, NAME2, NAME3, COLOR, MSKILL,\
    PARENT, MATERIAL, FLAGS, WEIGHT, SIZE, HARDNESS, HP, AMMO, BURST, TOHIT,\
    MELEE, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kWeapon, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kEnhanceable, NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0,\
    TOHIT, AMMO, BURST, MSKILL, kNoSkillCode, MELEE, NULL, NULL, NULL,\
    NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_MISSILE_WEAPON(enum, NAME1, NAME2, NAME3, COLOR, PARENT, MATERIAL,\
    FLAGS, WEIGHT, SIZE, HARDNESS, HP, TOHIT, MISSILE, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kWeapon, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kEnhanceable, NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0,\
    TOHIT, kObjNothing, 0, kNoSkillCode, kNoSkillCode, NULL, MISSILE, NULL,\
    NULL, NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_GUN(enum, NAME1, NAME2, NAME3, COLOR, GSKILL, PARENT, MATERIAL,\
    FLAGS, WEIGHT, SIZE, HARDNESS, HP, AMMO, BURST, TOHIT, GUN, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kWeapon, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kEnhanceable, NOINTRINSICS, RARITY, WEIGHT, SIZE, HARDNESS, HP, 0,\
    TOHIT, AMMO, BURST, kNoSkillCode, GSKILL, NULL, NULL, GUN, NULL,\
    NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

#define NEW_AMMUNITION(enum, NAME1, NAME2, NAME3, COLOR, PARENT, MATERIAL,\
    FLAGS, WEIGHT, SIZE, COST, RARITY) \
AllIlks[enum] = new shObjectIlk\
(\
    enum, kProjectile, PARENT, NAME1, NAME3, NAME2, COLOR, MATERIAL, COST,\
    FLAGS | kBugProof, NOINTRINSICS, RARITY, WEIGHT, SIZE, 1, 1, 0,\
    NOT_A_WEAPON, NOT_AN_IMPLANT, NOT_AN_ARMOR, 0, NOUSAGES\
)

/* Shorthand for buckazoids. Use to flag cost in object definition. */
#define BZ
/* Probability value for abstract ilks. */
#define ABSTRACT -1

#endif
